#include "src/core/file-controller.hpp"

#include "src/polyfill/base-directory.hpp"
#include "src/core/preset-controllers.hpp"
#include "src/core/hotkeys.hpp"
#include "src/ui/error-notifier.hpp"
#include "src/rhdc/core/rhdc-settings.hpp"
#include <fstream>
#include <ios>

#ifdef _WIN32
#include "src/polyfill/windows/unicode.hpp"
#endif

template<class T> class FileCache {

	private:
	T m_value;
	const fs::path m_filePath;
	void (*const m_setter)(JsonWriter&, const T &);

	void commit() const {
		try {
			std::ofstream file( m_filePath, std::ios_base::out | std::ios_base::trunc );
#ifdef _WIN32
			char buffer[8096];
			file.rdbuf()->pubsetbuf( buffer, 8096 );
#endif
			JsonWriter writer( &file, true );
			m_setter( writer, m_value );
			file.flush();
#ifdef _WIN32
			file.rdbuf()->pubsetbuf( nullptr, 0 );
#endif
		} catch( const std::exception &exception ) {
			string msg = "Failed to save data to '"s + m_filePath.u8string() + "'.\n";
			if( dynamic_cast<const JsonWriterException*>( &exception ) ) {
				msg += "Failed to serialize data to JSON. Please report this bug!\nDetails: ";
			} else if( dynamic_cast<const fs::filesystem_error*>( &exception ) ) {
				msg += "Filesystem Error: ";
			} else if( dynamic_cast<const std::ios_base::failure*>( &exception ) ) {
				msg += "I/O Error: ";
			}
			msg += exception.what();
			ErrorNotifier::alert( msg );
		}
	}

	public:
	FileCache(
		const fs::path &filePath,
		const T &defaultValue,
		T (*getter)(const Json&),
		void (*setter)(JsonWriter&, const T &)
	) : m_filePath( filePath ), m_setter( setter ) {
		if( !fs::isRegularFileSafe( filePath ) ) {
			m_value = defaultValue;
		} else try {
			std::ifstream file( filePath );
#ifdef _WIN32
			char buffer[8096];
			file.rdbuf()->pubsetbuf( buffer, 8096 );
#endif
			m_value = getter( Json::parse( file ) );
		} catch( const std::exception &exception ) {
			string msg = "Failed to load data from '"s + m_filePath.u8string() + "'.\n";
			if( dynamic_cast<const JsonReaderException*>( &exception ) ) {
				msg += "Failed to parse JSON data. Please report this bug!\nDetails: ";
			} else if( dynamic_cast<const fs::filesystem_error*>( &exception ) ) {
				msg += "Filesystem Error: ";
			} else if( dynamic_cast<const std::ios_base::failure*>( &exception ) ) {
				msg += "I/O Error: ";
			}
			msg += exception.what();
			ErrorNotifier::alert( msg );
			m_value = defaultValue;
		}
	}

	~FileCache() {}

	inline const T &get() const {
		return m_value;
	}

	inline void set( const T &value ) {
		m_value = value;
		commit();
	}

	inline void set( T &&value ) {
		m_value = std::move( value );
		commit();
	}

};

template<class T> static std::vector<T> parseJsonArray( const Json &json ) {
	const JArray &jsonArray = json.array();

	std::vector<T> parsedArray;
	parsedArray.reserve( jsonArray.size() );
	for( const Json &item : jsonArray ) {
		parsedArray.push_back( JsonSerializer::parse<T>( item ) );
	}
	return parsedArray;
}

template<class T> static void serializeToJsonArray( JsonWriter &writer, const std::vector<T> &array ) {
	writer.writeArrayStart();
	for( const T &item : array ) {
		JsonSerializer::serialize<T>( writer, item );
	}
	writer.writeArrayEnd();
}

static std::map<string, ControllerProfile> parseProfiles( const Json &json ) {
	const JArray &jsonArray = json.array();

	std::map<string, ControllerProfile> profiles;
	for( const Json &profileJson : jsonArray ) {
		ControllerProfile profile = JsonSerializer::parse<ControllerProfile>( profileJson );
		profiles[profile.name] = profile;
	}
	profiles[DefaultProfile::Gamecube.name] = DefaultProfile::Gamecube;
	profiles[DefaultProfile::XBox360.name] = DefaultProfile::XBox360;
	profiles[DefaultProfile::Nintendo64.name] = DefaultProfile::Nintendo64;
	return profiles;
}

static void serializeProfiles( JsonWriter &writer, const std::map<string, ControllerProfile> &profiles ) {
	writer.writeArrayStart();
	for( const auto &i : profiles ) {
		if( !DefaultProfile::exists( i.first ) ) {
			JsonSerializer::serialize<ControllerProfile>( writer, i.second );
		}
	}
	writer.writeArrayEnd();
}

static std::map<Uuid, InputMode> parseInputModes( const Json &json ) {
	const JArray &jsonArray = json.array();

	std::map<Uuid, InputMode> inputModes;
	for( const Json &profileJson : jsonArray ) {
		InputMode inputMode = JsonSerializer::parse<InputMode>( profileJson );
		inputModes[inputMode.id] = inputMode;
	}
	inputModes[DefaultInputModes::Normal.id] = DefaultInputModes::Normal;
	inputModes[DefaultInputModes::DualAnalog.id] = DefaultInputModes::DualAnalog;
	inputModes[DefaultInputModes::GoldenEye.id] = DefaultInputModes::GoldenEye;
	inputModes[DefaultInputModes::Clone.id] = DefaultInputModes::Clone;
	return inputModes;
}

static void serializeInputModes( JsonWriter &writer, const std::map<Uuid, InputMode> &inputModes ) {
	writer.writeArrayStart();
	for( const auto &i : inputModes ) {
		JsonSerializer::serialize<InputMode>( writer, i.second );
	}
	writer.writeArrayEnd();
}

static HashMap<Uuid,string> parseMappings( const Json &json ) {
	const JArray &jsonArray = json.array();

	HashMap<Uuid,string> mappings;
	for( const Json &mappingJson : jsonArray ) {
		const Uuid deviceId = Uuid::parse( mappingJson["device"].get<string>() );
		mappings[deviceId] = mappingJson["profile"].get<string>();
	}
	return mappings;
}

static void serializeMappings( JsonWriter &writer, const HashMap<Uuid,string> &mappings ) {
	writer.writeArrayStart();
	for( const auto &i : mappings ) {
		writer.writeObjectStart();
		writer.writeProperty( "device", i.first.toString() );
		writer.writeProperty( "profile", i.second );
		writer.writeObjectEnd();
	}
	writer.writeArrayEnd();
}

static std::vector<int> parseHotkeys( const Json &json ) {
	if( json.array().size() != (size_t)Hotkey::NUM_HOTKEYS ) {
		throw std::range_error( "Array of hotkeys is the wrong size" );
	}

	std::vector<int> hotkeys;
	hotkeys.reserve( (size_t)Hotkey::NUM_HOTKEYS );
	for( const Json &hotkey : json.array() ) {
		hotkeys.push_back( hotkey.get<int>() );
	}
	return hotkeys;
}

static void serializeHotkeys( JsonWriter &writer, const std::vector<int> &hotkeys ) {
	writer.writeArrayStart();
	for( int hotkey : hotkeys ) {
		writer.writeNumber( (int64)hotkey );
	}
	writer.writeArrayEnd();
}

static string parseString( const Json &json ) {
	return json.getOrDefault<string>( "" );
}

static void serializeString( JsonWriter &writer, const string &str ) {
	writer.writeString( str );
}

static FileCache<AppSettings> &appSettingsCache() {
	static FileCache<AppSettings> s_cache = FileCache<AppSettings>(
		BaseDir::config() / "settings.cfg",
		AppSettings::Default,
		JsonSerializer::parse<AppSettings>,
		JsonSerializer::serialize<AppSettings>
	);
	return s_cache;
}

static FileCache<std::vector<RomSource>> &romSourcesCache() {
	static FileCache<std::vector<RomSource>> s_cache = FileCache<std::vector<RomSource>>(
		BaseDir::data() / "sources.json",
		std::vector<RomSource>(),
		parseJsonArray<RomSource>,
		serializeToJsonArray<RomSource>
	);
	return s_cache;
}

static FileCache<std::map<string, ControllerProfile>> &profilesCache() {
	static FileCache<std::map<string, ControllerProfile>> s_cache = FileCache<std::map<string, ControllerProfile>>(
		BaseDir::data() / "profiles.json",
		{
			{ DefaultProfile::Gamecube.name, DefaultProfile::Gamecube },
			{ DefaultProfile::XBox360.name, DefaultProfile::XBox360 },
			{ DefaultProfile::Nintendo64.name, DefaultProfile::Nintendo64 }
		},
		parseProfiles,
		serializeProfiles
	);
	return s_cache;
}

static FileCache<std::map<Uuid, InputMode>> &inputModesCache() {
	static FileCache<std::map<Uuid, InputMode>> s_cache = FileCache<std::map<Uuid, InputMode>>(
		BaseDir::data() / "input_modes.json",
		{
			{ DefaultInputModes::Normal.id, DefaultInputModes::Normal },
			{ DefaultInputModes::DualAnalog.id, DefaultInputModes::DualAnalog },
			{ DefaultInputModes::GoldenEye.id, DefaultInputModes::GoldenEye },
			{ DefaultInputModes::Clone.id, DefaultInputModes::Clone }
		},
		parseInputModes,
		serializeInputModes
	);
	return s_cache;
}

static FileCache<HashMap<Uuid,string>> &mappingsCache() {
	static FileCache<HashMap<Uuid,string>> s_cache = FileCache<HashMap<Uuid,string>>(
		BaseDir::data() / "devices.json",
		HashMap<Uuid,string>(),
		parseMappings,
		serializeMappings
	);
	return s_cache;
}


static FileCache<UiState> &uiStateCache() {
	static FileCache<UiState> s_cache = FileCache<UiState>(
		BaseDir::cache() / "ui-state.json",
		UiState::Default,
		JsonSerializer::parse<UiState>,
		JsonSerializer::serialize<UiState>
	);
	return s_cache;
}

static FileCache<std::vector<int>> &hotkeysCache() {
	static FileCache<std::vector<int>> s_cache = FileCache<std::vector<int>>(
		BaseDir::config() / "hotkeys.json",
		std::vector<int>( Hotkeys::Default, Hotkeys::Default + (int)Hotkey::NUM_HOTKEYS ),
		parseHotkeys,
		serializeHotkeys
	);
	return s_cache;
}

static FileCache<InstalledVersionsInfo> &installedVersionsCache() {
	static FileCache<InstalledVersionsInfo> s_cache = FileCache<InstalledVersionsInfo>(
		BaseDir::data() / "cores.json",
		InstalledVersionsInfo::Default,
		JsonSerializer::parse<InstalledVersionsInfo>,
		JsonSerializer::serialize<InstalledVersionsInfo>
	);
	return s_cache;
}

static FileCache<RhdcSettings> &rhdcSettingsCache() {
	static FileCache<RhdcSettings> s_cache = FileCache<RhdcSettings>(
		BaseDir::config() / "rhdc.cfg",
		RhdcSettings::Default(),
		JsonSerializer::parse<RhdcSettings>,
		JsonSerializer::serialize<RhdcSettings>
	);
	return s_cache;
}

static FileCache<LayoutManager> &layoutManagerCache() {
	static FileCache<LayoutManager> s_cache = FileCache<LayoutManager>(
		BaseDir::cache() / "layouts.json",
		LayoutManager(),
		JsonSerializer::parse<LayoutManager>,
		JsonSerializer::serialize<LayoutManager>
	);
	return s_cache;
}

static FileCache<string> &lastControllerProfileNameCache() {
	static FileCache<string> s_cache = FileCache<string>(
		BaseDir::cache() / "last_profile.json",
		"",
		parseString,
		serializeString
	);
	return s_cache;
}

const AppSettings &FileController::loadAppSettings() {
	return appSettingsCache().get();
}

void FileController::saveAppSettings( const AppSettings &settings ) {
	appSettingsCache().set( settings );
}

const std::vector<RomSource> &FileController::loadRomSources() {
	return romSourcesCache().get();
}

void FileController::saveRomSources( const std::vector<RomSource> &sources ) {
	romSourcesCache().set( sources );
}

const std::map<string, ControllerProfile> &FileController::loadControllerProfiles() {
	return profilesCache().get();
}

void FileController::saveControllerProfiles( const std::map<string, ControllerProfile> &profiles ) {
	profilesCache().set( profiles );
}

const HashMap<Uuid,string> &FileController::loadControllerMappings() {
	return mappingsCache().get();
}

void FileController::saveControllerMappings( const HashMap<Uuid,string> &mappings ) {
	mappingsCache().set( mappings );
}

const UiState &FileController::loadUiState() {
	return uiStateCache().get();
}

void FileController::saveUiState( const UiState &state ) {
	uiStateCache().set( state );
}

const std::vector<int> &FileController::loadHotkeys() {
	return hotkeysCache().get();
}

void FileController::saveHotkeys( const std::vector<int> &hotkeys ) {
	hotkeysCache().set( hotkeys );
}

const std::map<Uuid,InputMode> &FileController::loadInputModes() {
	return inputModesCache().get();
}

void FileController::saveInputModes( const std::map<Uuid,InputMode> &modes ) {
	inputModesCache().set( modes );
}

const InstalledVersionsInfo &FileController::loadInstalledVersions() {
	return installedVersionsCache().get();
}

void FileController::saveInstalledVersions( const InstalledVersionsInfo &versions ) {
	installedVersionsCache().set( versions );
}

const RhdcSettings &FileController::loadRhdcSettings() {
	return rhdcSettingsCache().get();
}

void FileController::saveRhdcSettings( const RhdcSettings &settings ) {
	rhdcSettingsCache().set( settings );
}

const LayoutManager &FileController::loadLayoutManager() {
	return layoutManagerCache().get();
}

void FileController::saveLayoutManager( const LayoutManager &layoutManager ) {
	layoutManagerCache().set( layoutManager );
}

const ControllerProfile &FileController::loadLastControllerProfile() {
	const string &profileName = lastControllerProfileNameCache().get();
	const std::map<string, ControllerProfile> &profiles = FileController::loadControllerProfiles();
	const auto profile = profiles.find( profileName );
	if( profile == profiles.end() ) {
		return DefaultProfile::XBox360;
	} else {
		return profile->second;
	}
}

void FileController::saveLastControllerProfile( const string &name ) {
	lastControllerProfileNameCache().set( name );
}
