#include "src/core/retroarch.hpp"

#include "src/polyfill/base-directory.hpp"
#include "src/core/preset-controllers.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/numeric-string.hpp"
#include "src/core/hotkeys.hpp"
#include "src/input/gamepad-controller.hpp"
#include "src/input/keyboard.hpp"
#include "src/flatpak/flatpak.hpp"
#include <map>
#include <queue>
#include <unordered_map>
#include <cstdio>
#include <fstream>
#include <regex>
#include <cstring>
#include <cassert>
#include <cmath>

#ifndef _WIN32

#include <QFile>
#include <cstdlib>

static inline string escapePath( const fs::path &path ) {
	string escapedPath = "";
	for( const char c : path.u8string() ) {
#ifdef _WIN32
		if( c == '\"' || c == '\\' ) {
#else
		if( c == '\"' || c == '\\' || c == '`' || c == '$' ) {
#endif
			escapedPath += '\\';
		}
		escapedPath += c;
	}
	return escapedPath;
}

#endif

static fs::path expand_path( const fs::path &path ) {
	const string pathStr = path.u8string();
	if( pathStr == ":" ) {
		return RetroArch::getBasePath();
	} else if( pathStr.length() >= 2 && pathStr[0] == ':' && (pathStr[1] == '/' || pathStr[1] == '\\' ) ) {
#ifdef _WIN32
		return RetroArch::getBasePath() / fs::path( &path.wstring().c_str()[2] );
#else
		return BaseDir::home() / fs::path( &path.c_str()[2] );
#endif
	} else return BaseDir::expandHome( path );
}

static inline bool stringEndsWith( const string &str, const string &suffix ) {
	if( str.length() < suffix.length() ) return false;
	return std::strcmp( &str.c_str()[str.length() - suffix.length()], suffix.c_str() ) == 0;
}

static const char* s_plugins[] = {
	"auto",
	"parallel",
	"glide64",
	"angrylion",
	"gln64",
	"rice"
};

ubyte RetroArch::resolveUpscaling(
	ParallelUpscaling requestedUpscaling,
	ubyte windowScale
) noexcept {
	if( requestedUpscaling != ParallelUpscaling::Auto ) return 1 << ((ubyte)requestedUpscaling - 1);
	if( windowScale <= 1 ) return 1;
	if( windowScale == 2 ) return 2;
	if( windowScale < 8 ) return 4;
	return 8;
}

static const std::regex s_cfgRegex(
	"^([^=]+)\\s*=\\s*\"([^\"]*)\"",
	std::regex_constants::ECMAScript | std::regex_constants::optimize
);

static void loadConfigHelper( const fs::path &configPath, std::map<string,string> &configs ) {
	if( !fs::existsSafe( configPath ) ) return;
	std::ifstream configFile( configPath );
	while( configFile.good() && !configFile.eof() ) {
		string line;
		std::getline( configFile, line );

		std::smatch matches;
		if( std::regex_search( line, matches, s_cfgRegex ) ) {
			string key = matches[1];
			while( !key.empty() && key.back() == ' ') key.pop_back();
			if( !key.empty() ) {
				configs[key] = matches[2];
			}
		}
	}
}

static void loadConfig( std::map<string, string> &configs ) {
	loadConfigHelper( BaseDir::data() / "retroarch.cfg", configs );
	loadConfigHelper( BaseDir::data() / "retroarch-core-options.cfg", configs );
	const fs::path basePath = RetroArch::getBasePath();
	configs["libretro_directory"] = basePath.u8string();
	configs["savefile_directory"] = (basePath / "saves").u8string();
	configs["savestate_directory"] = (basePath / "states").u8string();
#ifdef _WIN32
	configs["assets_directory"] = (BaseDir::data() / "retroarch" / "assets").u8string();
#else
	configs["assets_directory"] = (BaseDir::data() / "retroarch" / "app" / "org.libretro.RetroArch" / "current" / "active" / "files" / "share" / "libretro" / "assets").u8string();
#endif
}

static void saveConfig( const std::map<string,string> &configs ) {
	std::ofstream mainConfigFile( BaseDir::data() / "retroarch.cfg", std::ios_base::out | std::ios_base::trunc );
	std::ofstream coreConfigFile( BaseDir::data() / "retroarch-core-options.cfg", std::ios_base::out | std::ios_base::trunc );

	for( const auto &i : configs ) {
		const string &key = i.first;
		const string &value = i.second;

		if(
			std::strncmp( key.c_str(), "parallel-n64-", sizeof( "parallel-n64-" ) - 1 ) == 0 ||
			std::strncmp( key.c_str(), "mupen64plus-", sizeof( "mupen64plus-" ) - 1 ) == 0
		) {
			coreConfigFile << key << " = \"" << value << '"' << std::endl;
		} else {
			mainConfigFile << key << " = \"" << value << '"' << std::endl;
		}
	}

	mainConfigFile << std::endl << std::flush;
	coreConfigFile << std::endl << std::flush;
}

static const char *s_inputNames[] = {
	"l_y_minus", // AnalogUp
	"l_y_plus", // AnalogDown
	"l_x_minus", // AnalogLeft
	"l_x_plus", // AnalogRight
	"r_y_minus", // CUp
	"r_y_plus", // CDown
	"r_x_minus", // CLeft
	"r_x_plus", // CRight
	"up", // DPadUp
	"down", // DPadDown
	"left", // DPadLeft
	"right", // DPadRight
	"b", // ButtonA
	"y", // ButtonB
	"l", // TriggerL
	"l2", // TriggerZ
	"r", // TriggerR
	"start" // Start
};

static void bindInput( std::map<string,string> &configs, const string &key, const Binding &binding ) {
	const string buttonKey = key + "_btn";
	const string axisKey = key + "_axis";

	// 99 and +99 are used instead of nul to prevent default bindings from messing things up
	switch( binding.type ) {
		case BindingType::None:
			configs[buttonKey] = "99";
			configs[axisKey] = "+99";
			break;
		case BindingType::Button:
			configs[buttonKey] = Number::toString( binding.buttonOrAxis );
			configs[axisKey] = "+99";
			break;
		case BindingType::AxisNegative:
			configs[buttonKey] = "99";
			configs[axisKey] = "-"s + Number::toString( binding.buttonOrAxis );
			break;
		case BindingType::AxisPositive:
			configs[buttonKey] = "99";
			configs[axisKey] = "+"s + Number::toString( binding.buttonOrAxis );
			break;
		case BindingType::HatUp:
			configs[buttonKey] = "h"s + Number::toString( binding.buttonOrAxis ) + "up";
			configs[axisKey] = "+99";
			break;
		case BindingType::HatDown:
			configs[buttonKey] = "h"s + Number::toString( binding.buttonOrAxis ) + "down";
			configs[axisKey] = "+99";
			break;
		case BindingType::HatLeft:
			configs[buttonKey] = "h"s + Number::toString( binding.buttonOrAxis ) + "left";
			configs[axisKey] = "+99";
			break;
		case BindingType::HatRight:
			configs[buttonKey] = "h"s + Number::toString( binding.buttonOrAxis ) + "right";
			configs[axisKey] = "+99";
			break;
	}
}

static inline HashMap<Uuid,std::queue<int>> getDevicePorts( int &numMapped ) {
	HashMap<Uuid,std::queue<int>> devicePorts;
	numMapped = 0;

	/* On Windows, lsjs needs to be in the same directory as the libraries it depends on.
	 * On Linux, it can be anywhere; however, if RetroArch is running in a Flatpak sandbox,
	 * then lsjs needs to be somewhere that sandbox can access.
	 *
	 * On Windows builds, lsjs is installed to the same directory as the main exe by the
	 * installer and just runs there. On Linux, I instead embed this helper program, then
	 * write it to a temporary folder accessible by Flatpak sandboxes.
	 */

	bool lsjsFailed = false;
	try {
#ifdef _WIN32
		const string lsjsCmd = (BaseDir::program() / "parallel-launcher-lsjs.exe").u8string();
#else
		const fs::path lsjsPath = BaseDir::homeTemp() / "lsjs";
		if( !fs::existsSafe( lsjsPath ) ) {
			QFile lsjsBin( ":/parallel-launcher-lsjs" );
			lsjsBin.open( QFile::ReadOnly );
			QByteArray lsjsData = lsjsBin.readAll();

			QFile lsjsTarget( lsjsPath.c_str() );
			lsjsTarget.open( QFile::WriteOnly | QFile::Truncate );
			lsjsTarget.setPermissions( QFileDevice::ReadOwner | QFileDevice::ExeOwner | QFileDevice::ReadGroup | QFileDevice::ExeGroup | QFileDevice::ReadOther | QFileDevice::ExeOther );
			lsjsTarget.write( lsjsData );
		}

		const string lsjsCmd = "flatpak run --installation=parallel-launcher --filesystem=home --command=\""s + escapePath( lsjsPath ) + "\" org.libretro.RetroArch";
#endif

		string uuidList;
		lsjsFailed = !Process::tryGetOutput( lsjsCmd, uuidList );
		if( !lsjsFailed ) {

			size_t lineStart = 0;
			for( int i = 0; lineStart < uuidList.length(); i++ ) {
				size_t lineEnd = uuidList.find( '\n', lineStart );
				if( lineEnd == string::npos ) {
					lineEnd = uuidList.length();
				}

				Uuid uuid;
				if( Uuid::tryParse( uuidList.substr( lineStart, lineEnd - lineStart ), uuid ) ) {
					devicePorts[uuid].push( i );
					numMapped++;
				}

				lineStart = lineEnd + 1;
			}
		}
	} catch( ... ) {
		lsjsFailed = true;
	}

	if( lsjsFailed ) {
		std::cerr << "Failed to run parallel-launcher-lsjs. Falling back to default port ordering. Controllers may not be bound correctly." << std::endl << std::flush;

		const std::vector<ConnectedGamepad> fallbackPortOrder = GamepadController::instance().getConnected();
		devicePorts.clear();
		devicePorts.reserve( fallbackPortOrder.size() );

		numMapped = (int)fallbackPortOrder.size();
		for( size_t i = 0; i < fallbackPortOrder.size(); i++ ) {
			devicePorts[fallbackPortOrder[i].info.uuid].push( (int)i );
		}
	}

	return devicePorts;
}

static inline string getAngrylionFilteringString( AngrylionFiltering filter ) {
	switch( filter ) {
		case AngrylionFiltering::Filter: return "Filtered";
		case AngrylionFiltering::Blur: return "AA+Blur";
		case AngrylionFiltering::Dedither: return "AA+Dedither";
		case AngrylionFiltering::AntiAlias: return "AA only";
		default: return "Unfiltered";
	}
}

static inline InputMode getInputMode( const Uuid &id ) {
	const std::map<Uuid,InputMode> &inputModes = FileController::loadInputModes();
	if( inputModes.count( id ) > 0 ) {
		return inputModes.at( id );
	}
	return DefaultInputModes::Normal;
}

AsyncProcess RetroArch::launchRom(
	fs::path romPath,
	const AppSettings &settings,
	const std::vector<PlayerController> &players,
	const RomInfo &romInfo,
	bool bindSavestate,
	GfxPlugin *pluginOut
) {
	std::map<string,string> configs;
	loadConfig( configs );

	EmulatorCore emulator = romInfo.emulator;
	if( emulator == EmulatorCore::UseDefault ) {
		emulator = settings.defaultEmulator;
	}

	GfxPlugin plugin;
	if( emulator == EmulatorCore::ParallelN64 ) {
		plugin = romInfo.parallelPlugin;
		if( plugin == GfxPlugin::UseDefault ) {
			plugin = settings.defaultParallelPlugin;
		}
	} else {
		plugin = romInfo.mupenPlugin;
		if( plugin == GfxPlugin::UseDefault ) {
			plugin = settings.defaultMupenPlugin;
		}
	}

	if( pluginOut != nullptr ) {
		*pluginOut = plugin;
	}

	const string width = Number::toString( 320 * (int)settings.windowScale );
	const string height = Number::toString( 240 * (int)settings.windowScale );
	const string widescreenWidth = Number::toString( (int)std::round( (double)settings.windowScale * 1280.0 / 3.0 ) );
	assert( emulator != EmulatorCore::UseDefault );
	assert( plugin != GfxPlugin::UseDefault );

	if( romInfo.overclockCPU || romInfo.overclockVI ) {
		configs["video_frame_delay"] = "0";
	}

	configs["video_fullscreen"] = "false";
	configs["suspend_screensaver_enable"] = "true";
	configs["video_force_aspect"] = romInfo.widescreen ? "false" : "true";
	configs["video_window_save_positions"] = "true";
	configs["parallel-n64-cpucore"] = "dynamic_recompiler";
	configs["mupen64plus-cpucore"] = "dynamic_recompiler";
	configs["parallel-n64-gfxplugin-accuracy"] = "veryhigh";
	configs["parallel-n64-parallel-rdp-native-texture-lod"] = "enabled";
	configs["mupen64plus-parallel-rdp-native-texture-lod"] = "True";
	configs["global_core_options"] = "true";
	configs["config_save_on_exit"] = "true";
	configs["save_file_compression"] = "false";
	configs["video_shader_enable"] = "false";
	configs["mupen64plus-ForceDisableExtraMem"] = "False";
	configs["savefiles_in_content_dir"] = "false";

	configs["video_scale_integer"] = (plugin == GfxPlugin::ParaLLEl && !romInfo.widescreen) ? "true" : "false";
	configs["video_vsync"] = settings.vsync ? "true" : "false";
	configs["video_windowed_position_width"] = configs["custom_viewport_width"] = romInfo.widescreen ? widescreenWidth : width;
	configs["video_windowed_position_height"] = configs["custom_viewport_height"] = height;
	configs["aspect_ratio_index"] = romInfo.widescreen ? "1" : "22";
	configs["mupen64plus-aspect"] = romInfo.widescreen ? "16:9" : "4:3";
	configs["parallel-n64-aspectratiohint"] = romInfo.widescreen ? "widescreen" : "normal";
	configs["parallel-n64-screensize"] = width + "x" + height;
	configs["mupen64plus-43screensize"] = width + "x" + height;
	configs["mupen64plus-169screensize"] = widescreenWidth + "x" + height;
	configs["parallel-n64-framerate"] = romInfo.overclockCPU ? "fullspeed" : "original";
	configs["mupen64plus-Framerate"] = romInfo.overclockCPU ? "Fullspeed" : "Original";
	configs["mupen64plus-CountPerOp"] = romInfo.overclockCPU ? "1" : "2";
	configs["parallel-n64-virefresh"] = romInfo.overclockVI ? "2200" : "auto";
	configs["mupen64plus-virefresh"] = romInfo.overclockVI ? "2200" : "Auto";
	configs["parallel-n64-parallel-rdp-upscaling"] = Number::toString( (int)resolveUpscaling( settings.parallelUpscaling, settings.windowScale ) ) + "x";
	configs["mupen64plus-parallel-rdp-upscaling"] = Number::toString( (int)resolveUpscaling( settings.parallelUpscaling, settings.windowScale ) ) + "x";
	configs["parallel-n64-parallel-rdp-vi-aa"] = configs["parallel-n64-parallel-rdp-vi-bilinear"] = settings.parallelAntiAliasing ? "enabled" : "disabled";
	configs["mupen64plus-parallel-rdp-vi-aa"] = configs["mupen64plus-parallel-rdp-vi-bilinear"] = settings.parallelAntiAliasing ? "True" : "False";
	configs["parallel-n64-angrylion-vioverlay"] = configs["mupen64plus-angrylion-vioverlay"] = getAngrylionFilteringString( settings.angrylionFiltering );
	configs["parallel-n64-parallel-rdp-native-tex-rect"] = romInfo.parallelTexRectUpscaling ? "disabled" : "enabled";
	configs["mupen64plus-parallel-rdp-native-tex-rect"] = romInfo.parallelTexRectUpscaling ? "False" : "True";
	configs["mupen64plus-BilinearMode"] = settings.glidenThreePointFiltering ? "3point" : "standard";
	configs["mupen64plus-EnableFBEmulation"] = romInfo.glidenFramebufferEmulation ? "True" : "False";
	configs["mupen64plus-EnableN64DepthCompare"] = (romInfo.glidenCorrectDepthCompare && romInfo.glidenFramebufferEmulation) ? "Compatible" : "False";
	configs["mupen64plus-FXAA"] = (settings.glidenAntiAliasing && romInfo.glidenFramebufferEmulation) ? "1" : "0";
	configs["mupen64plus-MultiSampling"] = (settings.glidenAntiAliasing && !romInfo.glidenFramebufferEmulation) ? "4" : "0";
	configs["parallel-n64-gfxplugin"] = s_plugins[(int)plugin];
	configs["mupen64plus-rdp-plugin"] = s_plugins[(int)plugin];
	configs["mupen64plus-rsp-plugin"] = (plugin == GfxPlugin::GlideN64) ? "hle" : "parallel";
	configs["parallel-n64-filtering"] = "automatic";
	configs["mupen64plus-EnableLODEmulation"] = "True";
	configs["mupen64plus-EnableCopyAuxToRDRAM"] = "True";
	configs["input_autodetect_enable"] = "false";
	configs["game_specific_options"] = "false";

	configs["parallel-n64-astick-deadzone"] = "0";
	configs["parallel-n64-astick-sensitivity"] = "100";
	configs["mupen64plus-astick-deadzone"] = "0";
	configs["mupen64plus-astick-sensitivity"] = "100";

	char floatStr[10];
	std::snprintf( floatStr, sizeof( floatStr ), "%0.6f", players.at( 0 ).profile.deadzone );
	configs["input_analog_deadzone"] = string( floatStr );
	std::snprintf( floatStr, sizeof( floatStr ), "%0.6f", players.at( 0 ).profile.sensitivity );
	configs["input_analog_sensitivity"] = string( floatStr );


	const InputMode inputMode = getInputMode( romInfo.inputModeId );
	const bool multiPort = inputMode.usesTwoPorts();

	const string throawayPath = (BaseDir::data() / ".temp").u8string();
	configs["input_remap_binds_enable"] = "true";
	configs["input_remapping_directory"] = throawayPath;
	configs["joypad_autoconfig_dir"] = throawayPath;
	configs["rgui_config_directory"] = throawayPath;
	configs["auto_remaps_enable"] = "false";
	configs["input_joypad_driver"] = "sdl2";
	configs["input_max_users"] = Number::toString( (int)players.size() * (multiPort ? 2 : 1) );

	std::error_code err;
	fs::remove_all( BaseDir::data() / ".temp", err );

	// Stub out controller bindings for unused ports
	for( char i = '1'; i <= '4'; i++ ) {
		configs["parallel-n64-pak"s + i] = "none";
		configs["mupen64plus-pak"s + i] = "none";

		const string prefix = "input_player"s + i + '_';
		const Binding nullBinding{ BindingType::None, 0 };
		configs[prefix + "joypad_index"] = "nul";

		for( int j = 0; j <= (int)ControllerAction::Start; j++ ) {
			bindInput( configs, prefix + s_inputNames[j], nullBinding );
		}
		bindInput( configs, prefix + "r2", nullBinding );
		bindInput( configs, prefix + "turbo", nullBinding );
	}

	int fallbackPort;
	HashMap<Uuid,std::queue<int>> devicePorts = getDevicePorts( fallbackPort );
	bool enableRumble = false;
	for( int i = 0; i < (int)players.size(); i++ ) {
		const char pc1 = '1' + (char)(multiPort ? 2*i : i);
		const char pc2 = pc1 + (char)1;

		const string indexKey1 = "input_player"s + pc1 + "_joypad_index";
		const string indexKey2 = "input_player"s + pc2 + "_joypad_index";
		if( !devicePorts[players[i].deviceUuid].empty() ) {
			const string deviceIndex = Number::toString( devicePorts.at( players[i].deviceUuid ).front() );
			configs[indexKey1] = deviceIndex;
			if( multiPort ) {
				configs[indexKey2] = deviceIndex;
			}
			devicePorts.at( players[i].deviceUuid ).pop();
		} else {
			const string deviceIndex = Number::toString( fallbackPort++ );
			configs[indexKey1] = deviceIndex;
			if( multiPort ) {
				configs[indexKey2] = deviceIndex;
			}
		}

		const string rumbleKey1 = "parallel-n64-pak"s + pc1;
		const string rumbleKey2 = "mupen64plus-pak"s + pc1;
		configs[rumbleKey1] = players[i].profile.rumble ? "rumble" : "memory";
		configs[rumbleKey2] = players[i].profile.rumble ? "rumble" : "memory";
		enableRumble |= players[i].profile.rumble;

		const Binding *bindings = players[i].profile.bindings;
		const Binding nullBinding = { BindingType::None, 0 };
		for( int j = 0; j <= (int)ControllerAction::Start; j++ ) {
			const string key = "input_player"s + pc1 + '_' + s_inputNames[j];
			const ControllerAction vi[2] = { inputMode.port1[j], inputMode.port2[j] };

			bindInput( configs, key, (vi[0] == ControllerAction::None) ? nullBinding : bindings[(ubyte)vi[0]] );
			if( multiPort ) {
				const string key = "input_player"s + pc2 + '_' + s_inputNames[j];
				bindInput( configs, key, (vi[1] == ControllerAction::None) ? nullBinding : bindings[(ubyte)vi[1]] );
			}
		}

		bindInput( configs, "input_player"s + pc1 + "_r2"s, { BindingType::None, 0 } );
		bindInput( configs, "input_player"s + pc1 + "_turbo"s, { BindingType::None, 0 } );
		if( multiPort ) {
			bindInput( configs, "input_player"s + pc2 + "_r2"s, { BindingType::None, 0 } );
			bindInput( configs, "input_player"s + pc2 + "_turbo"s, { BindingType::None, 0 } );
		}
	}

	if( bindSavestate ) {
		bindInput( configs, "input_load_state", players.at( 0 ).profile.bindings[(int)ControllerAction::LoadState] );
		bindInput( configs, "input_save_state", players.at( 0 ).profile.bindings[(int)ControllerAction::SaveState] );
		bindInput( configs, "input_toggle_fast_forward", players.at( 0 ).profile.bindings[(int)ControllerAction::FastForward] );
		bindInput( configs, "input_toggle_slowmotion", players.at( 0 ).profile.bindings[(int)ControllerAction::SlowMotion] );
	} else {
		configs["input_load_state_btn"] = "nul";
		configs["input_load_state_axis"] = "nul";
		configs["input_save_state_btn"] = "nul";
		configs["input_save_state_axis"] = "nul";
	}

	configs["enable_device_vibration"] = enableRumble ? "true" : "false";

	configs["input_desktop_menu_toggle"] = "nul";
	configs["input_grab_mouse_toggle"] = "nul";
	configs["input_game_focus_toggle"] = "nul";
	configs["input_send_debug_info"] = "nul";
	configs["input_shader_next"] = "nul";
	configs["input_shader_prev"] = "nul";
	configs["input_netplay_game_watch"] = "nul";
	configs["input_netplay_host_toggle"] = "nul";
	configs["input_osk_toggle"] = "nul";

	const std::vector<int> hotkeys = FileController::loadHotkeys();
	assert( hotkeys.size() == (size_t)Hotkey::NUM_HOTKEYS );
	for( ubyte i = 0; i < (ubyte)Hotkey::NUM_HOTKEYS; i++ ) {
		configs[Hotkeys::ConfigNames[i]] = Keycode::getNames( hotkeys[i] ).retroConfigName;
	}

	saveConfig( configs );

	const char *const coreName = (emulator == EmulatorCore::ParallelN64) ? "parallel_n64_libretro" : "mupen64plus_next_libretro";
	fs::path corePath = RetroArch::getCorePath() / coreName;

#ifdef _WIN32
	corePath += ".dll";
#else
	corePath += ".so";
#endif

	std::vector<string> args;
	args.reserve( 7 );
#ifndef _WIN32
	args.insert( args.end(), { "run"s, "--installation=parallel-launcher"s, "-p"s, "org.libretro.RetroArch"s });
#endif
	args.insert( args.end(), { "-L"s, corePath.u8string(), "--config"s, (BaseDir::data() / "retroarch.cfg").u8string(), romPath.u8string() } );

	return AsyncProcess(
#ifdef _WIN32
		RetroArch::getExePath().u8string(),
#else
		"flatpak",
#endif
		args
	);
}

fs::path RetroArch::getSaveFilePath( const fs::path &romPath ) {
	const fs::path saveDir = RetroArch::getBasePath() / "saves";
	if( saveDir.empty() ) {
		fs::path saveFilePath = expand_path( romPath );
		saveFilePath.replace_extension( ".srm" );
		return saveFilePath;
	} else {
		fs::path saveFilePath = expand_path( saveDir / romPath.filename() );
		saveFilePath.replace_extension( ".srm" );
		return saveFilePath;
	}

}

bool RetroArch::isRetroArchInstalled() {
#ifdef _WIN32
	return fs::existsSafe( RetroArch::getExePath() );
#else
	return Flatpak::isRetroArchInstalled();
#endif
}

static bool isCoreInstalled( const char *core ) {
	fs::path corePath = RetroArch::getCorePath();
	if( corePath.empty() ) return false;

	corePath = corePath / core;
#ifdef _WIN32
	corePath += ".dll";
#else
	corePath += ".so";
#endif

	return fs::isRegularFileSafe( corePath );
}

bool RetroArch::isEmulatorCoreInstalled( EmulatorCore core ) {
	switch( core ) {
		case EmulatorCore::ParallelN64: return isCoreInstalled( "parallel_n64_libretro" );
		case EmulatorCore::Mupen64plusNext: return isCoreInstalled( "mupen64plus_next_libretro" );
		default: return false;
	}
}

static const std::regex s_versionRegex(
	"^RetroArch.*v(\\d+)\\.(\\d+)(?:\\.\\d+)*(?:\\s|$)",
	std::regex_constants::ECMAScript | std::regex_constants::optimize | std::regex_constants::icase
);

bool RetroArch::resetConfig() {
	const fs::path mainConfigPath = BaseDir::data() / "retroarch.cfg";
	const fs::path coreConfigPath = BaseDir::data() / "retroarch-core-options.cfg";

	std::error_code err;
	if( fs::existsSafe( mainConfigPath ) ) fs::remove( mainConfigPath, err );
	if( fs::existsSafe( coreConfigPath ) ) fs::remove( coreConfigPath, err );
	return !err;
}

static fs::path createBasePath() {
	const fs::path basePath = BaseDir::data() / "retro-data";
	fs::create_directories( basePath );
	fs::create_directories( basePath / "saves" );
	fs::create_directories( basePath / "states" );
	fs::create_directories( basePath / "cores" );
	return basePath;
}

fs::path RetroArch::getBasePath() {
	static const fs::path s_path = createBasePath();
	return s_path;
}

fs::path RetroArch::getCorePath() {
	return RetroArch::getBasePath() / "cores";
}

fs::path RetroArch::getConfigPath() {
	return BaseDir::data();
}

#ifdef _WIN32
fs::path RetroArch::getExePath() {
	return BaseDir::data() / "retroarch" / "RetroArch.exe";
}
#endif

void RetroArch::migrateSaves() {
	std::map<string,string> configs;
	loadConfigHelper( BaseDir::data() / "retroarch.cfg", configs );
	std::error_code err;

	if( configs.count( "savefile_directory" ) > 0 ) {
		const fs::path oldSaveDir = expand_path( fs::to_path( configs["savefile_directory"] ) );
		if( fs::isDirectorySafe( oldSaveDir ) ) {
			const fs::path newSaveDir = RetroArch::getBasePath() / "saves";
			for( const auto &save : fs::directory_iterator( oldSaveDir ) ) {
				fs::copy( save.path(), newSaveDir / save.path().filename(), err );
			}
		}
	}

	if( configs.count( "savestate_directory" ) > 0 ) {
		const fs::path oldStateDir = expand_path( fs::to_path( configs["savestate_directory"] ) );
		if( fs::isDirectorySafe( oldStateDir ) ) {
			const fs::path newStateDir = RetroArch::getBasePath() / "states";
			for( const auto &state : fs::directory_iterator( oldStateDir ) ) {
				fs::copy( state.path(), newStateDir / state.path().filename(), err );
			}
		}
	}
}
