#include "src/core/settings.hpp"
#include "src/core/flags.hpp"
#include "src/core/filesystem.hpp"

const AppSettings AppSettings::Default = {
	/* visibleColumns */ RomInfoColumn::LastPlayed | RomInfoColumn::PlayTime,
	/* defaultParallelPlugin */ GfxPlugin::ParaLLEl,
	/* defaultMupenPlugin */ GfxPlugin::ParaLLEl,
	/* defaultEmulator */ EmulatorCore::ParallelN64,
	/* parallelUpscaling */ ParallelUpscaling::Auto,
	/* parallelAntiAliasing */ true,
	/* parallelTexRectUpscaling */ false,
	/* angrylionFiltering */ AngrylionFiltering::None,
	/* glidenAntiAliasing */ true,
	/* glidenThreePointFiltering */ true,
	/* glidenCorrectDepthCompare */ true,
	/* glidenFramebufferEmulation */ true,
	/* windowScale */ 4,
	/* vsync */ false,
	/* hideWhenPlaying */ false,
	/* preferredController */ std::nullopt,
#ifdef _WIN32
	/* windowsTheme */ "Fusion",
	/* checkForUpdates */ true,
#endif
	/* patchToSameFolder */ true,
	/* patchedRomFolder */ fs::to_path( "~/Documents/N64 Roms" ),
	/* migrationState */ 6,
	/* coreUpdateInterval */ CoreUpdateInterval::Daily,
	/* mupenDevBranch */ true
};

static constexpr char P_VISIBLE_COLUMNS[] = "visible_columns";
static constexpr char P_DEFAULT_PARALLEL_PLUGIN[] = "default_gfx_plugin";
static constexpr char P_DEFAULT_MUPEN_PLUGIN[] = "default_gfx_plugin_mupen";
static constexpr char P_DEFAULT_EMULATOR[] = "default_emulator";
static constexpr char P_PARALLEL_UPSCALING[] = "parallel_upscaling";
static constexpr char P_PARALLEL_AA[] = "parallel_antialiasing";
static constexpr char P_PARALLEL_UPSCALE_TEXRECT[] = "parallel_upscale_texrects";
static constexpr char P_ANGRYLION_FILTERING[] = "angrylion_filtering";
static constexpr char P_GLIDEN_AA[] = "gliden64_antialiasing";
static constexpr char P_GLIDEN_3POINT[] = "gliden64_3point_filtering";
static constexpr char P_GLIDEN_FIX_DEPTH[] = "gliden64_correct_depth_compare";
static constexpr char P_GLIDEN_EMU_FRAMEBUF[] = "gliden64_emulate_framebuffer";
static constexpr char P_WINDOW_SCALE[] = "window_scale";
static constexpr char P_VSYNC[] = "enable_vsync";
static constexpr char P_HIDE_WHILE_PLAYING[] = "hide_while_playing";
static constexpr char P_PREFERRED_CONTROLLER[] = "preferred_controller";
static constexpr char P_WINDOWS_THEME[] = "windows_theme";
static constexpr char P_CHECK_FOR_UPDATES[] = "check_for_updates";
static constexpr char P_PATCH_TO_SAME_FOLDER[] = "patch_to_same_folder";
static constexpr char P_PATCHED_ROM_FOLDER[] = "patched_rom_folder";
static constexpr char P_MIGRATION_STATE[] = "migration_state";
static constexpr char P_CORE_UPDATE_INTERVAL[] = "core_update_interval";
static constexpr char P_MUPEN_DEV[] = "use_mupen_dev_branch";

template<> void JsonSerializer::serialize<AppSettings>( JsonWriter &jw, const AppSettings &obj ) {
	jw.writeObjectStart();
	jw.writeProperty( P_VISIBLE_COLUMNS, obj.visibleColumns );
	jw.writeProperty( P_DEFAULT_PARALLEL_PLUGIN, obj.defaultParallelPlugin );
	jw.writeProperty( P_DEFAULT_MUPEN_PLUGIN, obj.defaultMupenPlugin );
	jw.writeProperty( P_DEFAULT_EMULATOR, obj.defaultEmulator );
	jw.writeProperty( P_PARALLEL_UPSCALING, obj.parallelUpscaling );
	jw.writeProperty( P_PARALLEL_AA, obj.parallelAntiAliasing );
	jw.writeProperty( P_PARALLEL_UPSCALE_TEXRECT, obj.parallelTexRectUpscaling );
	jw.writeProperty( P_ANGRYLION_FILTERING, obj.angrylionFiltering );
	jw.writeProperty( P_GLIDEN_AA, obj.glidenAntiAliasing );
	jw.writeProperty( P_GLIDEN_3POINT, obj.glidenThreePointFiltering );
	jw.writeProperty( P_GLIDEN_FIX_DEPTH, obj.glidenCorrectDepthCompare );
	jw.writeProperty( P_GLIDEN_EMU_FRAMEBUF, obj.glidenFramebufferEmulation );
	jw.writeProperty( P_WINDOW_SCALE, obj.windowScale );
	jw.writeProperty( P_VSYNC, obj.vsync );
	jw.writeProperty( P_HIDE_WHILE_PLAYING, obj.hideWhenPlaying );
	jw.writePropertyName( P_PREFERRED_CONTROLLER );
	if( obj.preferredController.has_value() ) {
		jw.writeString( obj.preferredController.value().toString() );
	} else {
		jw.writeNull();
	}
#ifdef _WIN32
	jw.writeProperty( P_WINDOWS_THEME, obj.windowsTheme );
	jw.writeProperty( P_CHECK_FOR_UPDATES, obj.checkForUpdates );
#endif
	jw.writeProperty( P_PATCH_TO_SAME_FOLDER, obj.patchToSameFolder );
	jw.writeProperty( P_PATCHED_ROM_FOLDER, obj.patchedRomFolder.u8string() );
	jw.writeProperty( P_MIGRATION_STATE, AppSettings::Default.migrationState );
	jw.writeProperty( P_CORE_UPDATE_INTERVAL, obj.coreUpdateInterval );
	jw.writeProperty( P_MUPEN_DEV, obj.mupenDevBranch );
	jw.writeObjectEnd();
}

template<> AppSettings JsonSerializer::parse<AppSettings>( const Json &json ) {
	return AppSettings{
		json[P_VISIBLE_COLUMNS].getOrDefault<RomInfoColumn>( AppSettings::Default.visibleColumns ),
		json[P_DEFAULT_PARALLEL_PLUGIN].getOrDefault<GfxPlugin>( AppSettings::Default.defaultParallelPlugin ),
		json[P_DEFAULT_MUPEN_PLUGIN].getOrDefault<GfxPlugin>( AppSettings::Default.defaultMupenPlugin ),
		json[P_DEFAULT_EMULATOR].getOrDefault<EmulatorCore>( AppSettings::Default.defaultEmulator ),
		json[P_PARALLEL_UPSCALING].getOrDefault<ParallelUpscaling>( AppSettings::Default.parallelUpscaling ),
		json[P_PARALLEL_AA].getOrDefault<bool>( AppSettings::Default.parallelAntiAliasing ),
		json[P_PARALLEL_UPSCALE_TEXRECT].getOrDefault<bool>( AppSettings::Default.parallelTexRectUpscaling ),
		json[P_ANGRYLION_FILTERING].getOrDefault<AngrylionFiltering>( AppSettings::Default.angrylionFiltering ),
		json[P_GLIDEN_AA].getOrDefault<bool>( AppSettings::Default.glidenAntiAliasing ),
		json[P_GLIDEN_3POINT].getOrDefault<bool>( AppSettings::Default.glidenThreePointFiltering ),
		json[P_GLIDEN_FIX_DEPTH].getOrDefault<bool>( AppSettings::Default.glidenCorrectDepthCompare ),
		json[P_GLIDEN_EMU_FRAMEBUF].getOrDefault<bool>( AppSettings::Default.glidenFramebufferEmulation ),
		json[P_WINDOW_SCALE].getOrDefault<ubyte>( AppSettings::Default.windowScale ),
		json[P_VSYNC].getOrDefault<bool>( AppSettings::Default.vsync ),
		json[P_HIDE_WHILE_PLAYING].getOrDefault<bool>( AppSettings::Default.hideWhenPlaying ),
		(json[P_PREFERRED_CONTROLLER].exists() && !json[P_PREFERRED_CONTROLLER].isNull()) ?
			Uuid::parse( json[P_PREFERRED_CONTROLLER].get<string>() ) : std::optional<Uuid>(),
#ifdef _WIN32
		json[P_WINDOWS_THEME].getOrDefault<string>( AppSettings::Default.windowsTheme ),
		json[P_CHECK_FOR_UPDATES].getOrDefault<bool>( AppSettings::Default.checkForUpdates ),
#endif
		json[P_PATCH_TO_SAME_FOLDER].getOrDefault<bool>( AppSettings::Default.patchToSameFolder ),
		fs::to_path( json[P_PATCHED_ROM_FOLDER].getOrDefault<string>( AppSettings::Default.patchedRomFolder.u8string() ) ),
		json[P_MIGRATION_STATE].getOrDefault<short>( 0 ),
		json[P_CORE_UPDATE_INTERVAL].getOrDefault<CoreUpdateInterval>( AppSettings::Default.coreUpdateInterval ),
		json[P_MUPEN_DEV].getOrDefault<bool>( AppSettings::Default.mupenDevBranch )
	};
}
