#include "src/db/data-provider.hpp"

#include "src/db/sqlite.hpp"
#include "src/core/special-groups.hpp"

void DataProvider::addRomInfo( const RomInfo &romInfo ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO ROMS(
	sha1,
	name,
	internal_name,
	emulator,
	parallel_plugin,
	mupen_plugin,
	upscale_texrects,
	emulate_framebuffer,
	n64_depth_compare,
	overclock_cpu,
	overclock_vi,
	widescreen,
	input_mode_id,
	last_played,
	play_time,
	crc32
) VALUES (
	?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15, ?16
)
)#"
	);

	s_cmd.addParameter( romInfo.sha1, DATA_REF );
	s_cmd.addParameter( romInfo.name, DATA_REF );
	s_cmd.addParameter( romInfo.internalName, DATA_REF );
	s_cmd.addParameter( (int)romInfo.emulator );
	s_cmd.addParameter( (int)romInfo.parallelPlugin );
	s_cmd.addParameter( (int)romInfo.mupenPlugin );
	s_cmd.addParameter( romInfo.parallelTexRectUpscaling );
	s_cmd.addParameter( romInfo.glidenFramebufferEmulation );
	s_cmd.addParameter( romInfo.glidenCorrectDepthCompare );
	s_cmd.addParameter( romInfo.overclockCPU );
	s_cmd.addParameter( romInfo.overclockVI );
	s_cmd.addParameter( romInfo.widescreen );
	s_cmd.addParameter( romInfo.inputModeId, DATA_REF );
	s_cmd.addParameter( romInfo.lastPlayed );
	s_cmd.addParameter( romInfo.playTime );
	s_cmd.addParameter( (int64)romInfo.crc32 );
	s_cmd.execNonQuery();

	for( const string &group : romInfo.groups ) {
		static SqlCommand s_cmd2( R"#(
INSERT OR IGNORE INTO ROM_GROUPS( sha1, group_name )
VALUES( ?1, ?2 )
)#"
		);

		s_cmd2.addParameter( romInfo.sha1, DATA_REF );
		s_cmd2.addParameter( group, DATA_REF );
		s_cmd2.execNonQuery();
	}
}

void DataProvider::addRomFile( const RomFile &romFile ) {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO ROM_PATHS( path, last_modified, sha1 )
VALUES( ?1, ?2, ?3 )
)#"
	);

	s_cmd.addParameter( romFile.path.u8string(), DATA_COPY );
	s_cmd.addParameter( romFile.lastModified );
	s_cmd.addParameter( romFile.sha1, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::addGroup( const string &groupName ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO PERSISTENT_GROUPS( name )
VALUES( ?1 )
)#"
	);

	s_cmd.addParameter( groupName, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::deleteGroup( const string &groupName ) noexcept {
	static SqlBatch s_cmd( R"#(
DELETE FROM PERSISTENT_GROUPS 
WHERE name = ?1;

DELETE FROM ROM_GROUPS
WHERE group_name = ?1;
)#"
	);

	s_cmd.addParameter( groupName, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::renameGroup( const string &oldName, const string &newName ) noexcept {
	static SqlBatch s_cmd( R"#(
UPDATE PERSISTENT_GROUPS
SET name = ?2
WHERE name = ?1;

UPDATE ROM_GROUPS
SET group_name = ?2
WHERE group_name = ?1;
)#"
	);

	s_cmd.addParameter( oldName, DATA_REF );
	s_cmd.addParameter( newName, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomEmulator( const string &sha1, EmulatorCore core ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET emulator = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( (int)core );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomParallelPlugin( const string &sha1, GfxPlugin plugin ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET parallel_plugin = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( (int)plugin );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomMupenPlugin( const string &sha1, GfxPlugin plugin ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET mupen_plugin = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( (int)plugin );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomCpuOverclocking( const string &sha1, bool overclock ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET overclock_cpu = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( overclock );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomViOverclocking( const string &sha1, bool overclock ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET overclock_vi = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( overclock );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomWidescreen( const string &sha1, bool widescreen ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET widescreen = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( widescreen );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomInputMode( const string &sha1, const Uuid &inputModeId ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET input_mode_id = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( inputModeId, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::updatePlayTime(
	const string &sha1,
	int64 lastPlayed,
	int64 playTime
) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS SET
	last_played = ?2,
	play_time = ?3
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( lastPlayed );
	s_cmd.addParameter( playTime );
	s_cmd.execNonQuery();
}

void DataProvider::renameRom( const string &sha1, const string &name ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET name = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( name, DATA_REF );
	s_cmd.execNonQuery();
}

static void addGroups( std::set<string> &groups, const string &groupsString ) {
	size_t nameStart = 0;
	for( size_t i = 0; i < groupsString.length(); i++ ) {
		if( groupsString[i] != '\n' ) continue;
		if( i != nameStart ) {
			groups.insert( groupsString.substr( nameStart, i - nameStart ) );
		}
		nameStart = i + 1;
	}

	if( nameStart < groupsString.length() ) {
		groups.insert( groupsString.substr( nameStart ) );
	}
}

bool DataProvider::tryFetchRomByHash( const string &sha1, bool includeGroups, RomInfo *romInfo ) noexcept {
	static SqlCommand s_cmdWithoutGroups( R"#(
SELECT name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32
FROM ROMS
WHERE sha1 = ?1
)#"
	);

	static SqlCommand s_cmdWithGroups( R"#(
SELECT name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32, groups
FROM TAGGED_ROMS
WHERE sha1 = ?1
)#"
	);

	SqlCommand &cmd = (includeGroups && romInfo != nullptr) ? s_cmdWithGroups : s_cmdWithoutGroups;
	cmd.addParameter( sha1, DATA_REF );

	DataRecordIterator results = cmd.execQuery();
	if( !results.moveNext() ) return false;

	if( romInfo != nullptr ) {
		romInfo->sha1 = sha1;
		romInfo->name = results.getString( 0 );
		romInfo->internalName = results.getString( 1 );
		romInfo->emulator = (EmulatorCore)results.getInt( 2 );
		romInfo->parallelPlugin = (GfxPlugin)results.getInt( 3 );
		romInfo->mupenPlugin = (GfxPlugin)results.getInt( 4 );
		romInfo->parallelTexRectUpscaling = results.getBool( 5 );
		romInfo->glidenFramebufferEmulation = results.getBool( 6 );
		romInfo->glidenCorrectDepthCompare = results.getBool( 7 );
		romInfo->overclockCPU = results.getBool( 8 );
		romInfo->overclockVI = results.getBool( 9 );
		romInfo->widescreen = results.getBool( 10 );
		romInfo->inputModeId = results.getUuid( 11 );
		romInfo->lastPlayed = results.getLong( 12 );
		romInfo->playTime = results.getLong( 13 );
		romInfo->crc32 = results.getUInt( 14 );
		romInfo->groups = std::set<string>();

		if( includeGroups && !results.isNull( 15 ) ) {
			addGroups( romInfo->groups, results.getString( 15 ) );
		}
	}

	return true;
}

bool DataProvider::tryFetchRomByPath( const fs::path &filePath, bool includeGroups, RomInfo *romInfo, RomFile *romFile ) {
	static SqlCommand s_cmdWithoutGroups( R"#(
SELECT sha1, name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32, last_modified
FROM ROMS INNER JOIN ROM_PATHS USING( sha1 )
WHERE path = ?1
)#"
	);

	static SqlCommand s_cmdWithGroups( R"#(
SELECT sha1, name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32, last_modified, groups
FROM TAGGED_ROMS INNER JOIN ROM_PATHS USING( sha1 )
WHERE path = ?1
)#"
	);

	SqlCommand &cmd = (includeGroups && romInfo != nullptr) ? s_cmdWithGroups : s_cmdWithoutGroups;
	cmd.addParameter( filePath.u8string(), DATA_COPY );

	DataRecordIterator results = cmd.execQuery();
	if( !results.moveNext() ) return false;

	if( romInfo != nullptr ) {
		romInfo->sha1 = results.getString( 0 );
		romInfo->name = results.getString( 1 );
		romInfo->internalName = results.getString( 2 );
		romInfo->emulator = (EmulatorCore)results.getInt( 3 );
		romInfo->parallelPlugin = (GfxPlugin)results.getInt( 4 );
		romInfo->mupenPlugin = (GfxPlugin)results.getInt( 5 );
		romInfo->parallelTexRectUpscaling = results.getBool( 6 );
		romInfo->glidenFramebufferEmulation = results.getBool( 7 );
		romInfo->glidenCorrectDepthCompare = results.getBool( 8 );
		romInfo->overclockCPU = results.getBool( 9 );
		romInfo->overclockVI = results.getBool( 10 );
		romInfo->widescreen = results.getBool( 11 );
		romInfo->inputModeId = results.getUuid( 12 );
		romInfo->lastPlayed = results.getLong( 13 );
		romInfo->playTime = results.getLong( 14 );
		romInfo->crc32 = results.getUInt( 15 );
		romInfo->groups = std::set<string>();

		if( includeGroups && !results.isNull( 17 ) ) {
			addGroups( romInfo->groups, results.getString( 17 ) );
		}
	}

	if( romFile != nullptr ) {
		romFile->path = filePath;
		romFile->lastModified = results.getLong( 16 );
		romFile->sha1 = results.getString( 0 );
	}

	return true;
}

bool DataProvider::tryFetchRomFile( const fs::path &filePath, RomFile *romFile ) {
	static SqlCommand s_cmd( R"#(
SELECT last_modified, sha1
FROM ROM_PATHS INNER JOIN ROMS USING( sha1 )
WHERE path = ?1
)#"
	);

	s_cmd.addParameter( filePath.u8string(), DATA_COPY );
	DataRecordIterator results = s_cmd.execQuery();
	if( results.moveNext() ) {
		if( romFile != nullptr ) {
			romFile->path = filePath;
			romFile->lastModified = results.getLong( 0 );
			romFile->sha1 = results.getString( 1 );
		}
		return true;
	} else return false;
}

bool DataProvider::tryFetchRomPathWithCrc32( uint crc32, fs::path &filePath ) {
	static SqlCommand s_cmd( R"#(
SELECT path
FROM ROMS INNER JOIN ROM_PATHS USING( sha1 )
WHERE crc32 = ?1
)#"
	);

	s_cmd.addParameter( crc32 );

	DataRecordIterator results = s_cmd.execQuery();
	while( results.moveNext() ) {
		filePath = results.getPath( 0 );
		if( fs::existsSafe( filePath ) ) return true;
	}

	return false;
}

bool DataProvider::tryFetchRomPathWithSha1( const string &sha1, fs::path &filePath ) {
	static SqlCommand s_cmd( R"#(
SELECT path
FROM ROMS INNER JOIN ROM_PATHS USING( sha1 )
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );

	DataRecordIterator results = s_cmd.execQuery();
	while( results.moveNext() ) {
		filePath = results.getPath( 0 );
		if( fs::existsSafe( filePath ) ) return true;
	}

	return false;
}

int DataProvider::countRomFilesWithSha1( const string &sha1 ) noexcept {
	static SqlCommand s_cmd( R"#(
SELECT COUNT(*)
FROM ROM_PATHS
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	DataRecordIterator results = s_cmd.execQuery();
	return results.moveNext() ? results.getInt( 0 ) : 0;
}

int DataProvider::countRomsInGroup( const string &groupName ) noexcept {
	static SqlCommand s_cmd( R"#(
SELECT COUNT(*)
FROM ROM_PATHS INNER JOIN ROM_GROUPS USING( sha1 )
WHERE group_name = ?1
)#"
	);

	s_cmd.addParameter( groupName, DATA_REF );
	DataRecordIterator results = s_cmd.execQuery();
	return results.moveNext() ? results.getInt( 0 ) : 0;
}

bool DataProvider::romWithSha1Exists( const string &sha1 ) noexcept {
	static SqlCommand s_cmd( R"#(
SELECT 1
FROM ROMS
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	return s_cmd.execQuery().moveNext();
}

bool DataProvider::romWithPathExists( const fs::path &filePath ) {
	static SqlCommand s_cmd( R"#(
SELECT 1
FROM ROMS INNER JOIN ROM_PATHS USING( sha1 )
WHERE path = ?1
)#"
	);

	s_cmd.addParameter( filePath.u8string(), DATA_COPY );
	return s_cmd.execQuery().moveNext();
}

std::set<string> DataProvider::fetchPersistentGroups() {
	static SqlCommand s_cmd( R"#(
SELECT name FROM PERSISTENT_GROUPS
ORDER BY name ASC
)#"
	);

	std::set<string> results;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		results.insert( iterator.getString( 0 ) );
	}

	return results;
}

std::vector<string> DataProvider::fetchAllGroups() {
	static SqlCommand s_cmd( R"#(
SELECT DISTINCT group_name AS name
FROM ROM_GROUPS INNER JOIN ROM_PATHS USING( sha1 )
UNION
SELECT name FROM PERSISTENT_GROUPS
ORDER BY name ASC
)#"
	);

	std::vector<string> results;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		results.push_back( iterator.getString( 0 ) );
	}

	return results;
}

std::vector<RomFile> DataProvider::fetchRomFiles() {
	static SqlCommand s_cmd( R"#(
SELECT path, last_modified, sha1
FROM ROM_PATHS INNER JOIN ROMS USING( sha1 )
)#"
	);

	std::vector<RomFile> results;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		results.push_back({
			iterator.getPath( 0 ),
			iterator.getLong( 1 ),
			iterator.getString( 2 )
		});
	}

	return results;
}

HashMap<string,RomInfo> DataProvider::fetchAllRomInfo() {
	static SqlCommand s_cmd( R"#(
SELECT sha1, name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32, groups
FROM TAGGED_ROMS
)#"
	);

	HashMap<string,RomInfo> results;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		string sha1 = iterator.getString( 0 );
		RomInfo &romInfo = results[sha1];
		romInfo.sha1 = std::move( sha1 );
		romInfo.name = iterator.getString( 1 );
		romInfo.internalName = iterator.getString( 2 );
		romInfo.emulator = (EmulatorCore)iterator.getInt( 3 );
		romInfo.parallelPlugin = (GfxPlugin)iterator.getInt( 4 );
		romInfo.mupenPlugin = (GfxPlugin)iterator.getInt( 5 );
		romInfo.parallelTexRectUpscaling = iterator.getBool( 6 );
		romInfo.glidenFramebufferEmulation = iterator.getBool( 7 );
		romInfo.glidenCorrectDepthCompare = iterator.getBool( 8 );
		romInfo.overclockCPU = iterator.getBool( 9 );
		romInfo.overclockVI = iterator.getBool( 10 );
		romInfo.widescreen = iterator.getBool( 11 );
		romInfo.inputModeId = iterator.getUuid( 12 );
		romInfo.lastPlayed = iterator.getLong( 13 );
		romInfo.playTime = iterator.getLong( 14 );
		romInfo.crc32 = iterator.getUInt( 15 );
		romInfo.groups = std::set<string>();
		if( !iterator.isNull( 16 ) ) {
			addGroups( romInfo.groups, iterator.getString( 16 ) );
		}
	}

	return results;
}

HashSet<string> DataProvider::fetchAllKnownChecksums() {
	static SqlCommand s_cmd( "SELECT DISTINCT sha1 FROM ROM_PATHS" );

	HashSet<string> checksums;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		checksums.insert( iterator.getString( 0 ) );
	}
	return checksums;
}

std::vector<fs::path> DataProvider::fetchManualRomPaths() {
	static SqlCommand s_cmd( R"#(
SELECT path
FROM MANUAL_ROM_PATHS
ORDER BY path ASC
)#" );

	std::vector<fs::path> results;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		results.push_back( iterator.getPath( 0 ) );
	}

	return results;
}

void DataProvider::addManualRomPath( const fs::path &path ) {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO MANUAL_ROM_PATHS( path )
VALUES( ?1 )
)#"
	);

	s_cmd.addParameter( path.u8string(), DATA_COPY );
	s_cmd.execNonQuery();
}

void DataProvider::deleteManualRomPath( const fs::path &path ) {
	static SqlCommand s_cmd( R"#(
DELETE FROM MANUAL_ROM_PATHS
WHERE path = ?1
)#"
	);

	s_cmd.addParameter( path.u8string(), DATA_COPY );
	s_cmd.execNonQuery();
}

void DataProvider::deleteAllManualRomPaths() noexcept {
	static SqlCommand s_cmd( "DELETE FROM MANUAL_ROM_PATHS" );
	s_cmd.execNonQuery();
}

void DataProvider::updateRomFile( const RomFile &romFile ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROM_PATHS SET
	last_modified = ?2,
	sha1 = ?3
WHERE path = ?1
)#"
	);

	s_cmd.addParameter( romFile.path.u8string(), DATA_COPY );
	s_cmd.addParameter( romFile.lastModified );
	s_cmd.addParameter( romFile.sha1, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::cloneRomInfo(
	const string &oldSha1,
	const string &newSha1,
	const string &internalName,
	uint crc32
) noexcept {
	static SqlBatch s_cmd( R"#(
INSERT OR IGNORE INTO ROMS(
	sha1, name, internal_name, emulator, parallel_plugin, mupen_plugin, upscale_texrects, emulate_framebuffer, n64_depth_compare, overclock_cpu, overclock_vi, widescreen, input_mode_id, last_played, play_time, crc32
)
SELECT
	?2 AS sha1,
	name,
	?3 AS internal_name,
	emulator,
	parallel_plugin,
	mupen_plugin,
	upscale_texrects,
	emulate_framebuffer,
	n64_depth_compare,
	overclock_cpu,
	overclock_vi,
	widescreen,
	input_mode_id,
	last_played,
	play_time,
	?4 AS crc32
FROM ROMS
WHERE sha1 = ?1 
ON CONFLICT( sha1 ) DO UPDATE SET
	internal_name = excluded.internal_name,
	crc32 = excluded.crc32,
	last_played = max( ROMS.last_played, excluded.last_played ),
	play_time = ROMS.play_time + excluded.play_time;

INSERT OR IGNORE INTO ROM_GROUPS( sha1, group_name )
SELECT ?2 AS sha1, group_name
FROM ROM_GROUPS
WHERE sha1 = ?1;
)#"
	);

	s_cmd.addParameter( oldSha1, DATA_REF );
	s_cmd.addParameter( newSha1, DATA_REF );
	s_cmd.addParameter( internalName, DATA_REF );
	s_cmd.addParameter( crc32 );
	s_cmd.execNonQuery();
}

void DataProvider::deleteRomPath( const fs::path &filePath ) {
	static SqlCommand s_cmd( R"#(
DELETE FROM ROM_PATHS
WHERE path = ?1
)#"
	);

	s_cmd.addParameter( filePath.u8string(), DATA_COPY );
	s_cmd.execNonQuery();
}

void DataProvider::deleteRomInfo( const string &sha1 ) noexcept {
	static SqlBatch s_cmd( R"#(
DELETE FROM ROMS
WHERE sha1 = ?1;

DELETE FROM ROM_GROUPS
WHERE sha1 = ?1;
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::addToGroup( const string &sha1, const string &groupName ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO ROM_GROUPS( sha1, group_name )
VALUES( ?1, ?2 )
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( groupName, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::removeFromGroup( const string &sha1, const string &groupName ) noexcept {
	static SqlCommand s_cmd( R"#(
DELETE FROM ROM_GROUPS
WHERE sha1 = ?1
AND group_name = ?2
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( groupName, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::setRomParallelTexRectUpscaling( const string &sha1, bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET upscale_texrects = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setRomGlidenFrameBufferEmulation( const string &sha1, bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET emulate_framebuffer = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setRomGlidenAccurateDepthCompare( const string &sha1, bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET n64_depth_compare = ?2
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setDefaultParallelTexRectUpscaling( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET upscale_texrects = ?1
	WHERE emulator == 0 
	OR (emulator == 1 AND parallel_plugin != 1)
	OR (emulator == 2 AND mupen_plugin != 1)
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setDefaultGlidenFrameBufferEmulation( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET emulate_framebuffer = ?1
	WHERE emulator != 2
	OR mupen_plugin != 2
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setDefaultGlidenAccurateDepthCompare( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET n64_depth_compare = ?1
	WHERE emulator != 2
	OR mupen_plugin != 2
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setAllParallelTexRectUpscaling( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET upscale_texrects = ?1
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setAllGlidenFrameBufferEmulation( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET emulate_framebuffer = ?1
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::setAllGlidenAccurateDepthCompare( bool enabled ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET n64_depth_compare = ?1
)#"
	);

	s_cmd.addParameter( enabled );
	s_cmd.execNonQuery();
}

void DataProvider::addOrUpdateRhdcHack( const RhdcHackInfo &hack ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR REPLACE INTO RHDC_HACKS( id, name, slug, description, star_count, downloads, avg_rating, avg_difficulty, category )
VALUES( ?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9 )
)#"
	);

	s_cmd.addParameter( hack.hackId, DATA_REF );
	s_cmd.addParameter( hack.name, DATA_REF );
	s_cmd.addParameter( hack.slug, DATA_REF );
	s_cmd.addParameter( hack.description, DATA_REF );
	s_cmd.addParameter( hack.starCount );
	s_cmd.addParameter( hack.downloads );
	s_cmd.addParameter( hack.avgRating );
	s_cmd.addParameter( hack.avgDifficulty );
	s_cmd.addParameter( hack.category, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::addRhdcHackVersion( const string &hackId, const string &sha1 ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO RHDC_HACK_VERSIONS( sha1, id )
VALUES( ?2, ?1 )
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( sha1, DATA_REF );
	s_cmd.execNonQuery();
}

void DataProvider::clearRhdcHacks() noexcept {
	static SqlBatch s_cmd( R"#(
		DELETE FROM RHDC_HACKS;
		DELETE FROM ROM_GROUPS WHERE group_name = ?1;
		DELETE FROM ROM_GROUPS WHERE group_name = ?2;
		DELETE FROM ROM_GROUPS WHERE group_name = ?3;
)#"
	);

	s_cmd.addParameter( SpecialGroups::WantToPlay, DATA_REF );
	s_cmd.addParameter( SpecialGroups::InProgress, DATA_REF );
	s_cmd.addParameter( SpecialGroups::Completed, DATA_REF );

	s_cmd.execNonQuery();
}

void DataProvider::updateRhdcProgress( const string &hackId, const RhdcHackProgress &progress ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR REPLACE INTO RHDC_HACK_PROGRESS( id, my_rating, my_difficulty, stars_collected, completed )
VALUES( ?1, ?2, ?3, ?4, ?5 )
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( progress.rating );
	s_cmd.addParameter( progress.difficulty );
	s_cmd.addParameter( progress.stars );
	s_cmd.addParameter( progress.completed );
	s_cmd.execNonQuery();
}

void DataProvider::updateRhdcHackRating( const string &hackId, ubyte quality, ubyte difficulty ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE RHDC_HACK_PROGRESS SET
	my_rating = ?2,
	my_difficulty = ?3
WHERE id = ?1
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( quality );
	s_cmd.addParameter( difficulty );
	s_cmd.execNonQuery();
}

void DataProvider::updateRhdcStarProgress( const string &hackId, ushort stars, bool completed ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE RHDC_HACK_PROGRESS SET
	stars_collected = ?2,
	completed = ?3
WHERE id = ?1
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( stars );
	s_cmd.addParameter( completed );
	s_cmd.execNonQuery();
}

void DataProvider::suggestRhdcHackPlayTime( const string &hackId, int64 playTime ) noexcept {
	static SqlCommand s_cmd( R"#(
UPDATE ROMS
SET play_time = ?2
WHERE ROMS.sha1 IN( SELECT sha1 FROM RHDC_HACK_VERSIONS WHERE id = ?1 )
AND play_time < ?2
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( playTime );
	s_cmd.execNonQuery();
}

HashMap<string,BasicRhdcHack> DataProvider::fetchFollowedRhdcHacks() {
	static SqlCommand s_cmd( R"#(
SELECT id, name, star_count, category, my_rating, my_difficulty, stars_collected, completed
FROM RHDC_HACKS LEFT JOIN RHDC_HACK_PROGRESS USING( id )
)#"
	);

	HashMap<string,BasicRhdcHack> hackMap;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		const string hackId = iterator.getString( 0 );
		hackMap[hackId] = BasicRhdcHack{
			BasicRhdcHackInfo {
				hackId,
				iterator.getString( 1 ),
				iterator.getInt( 2 ),
				iterator.getString( 3 ) == "Kaizo"
			},
			iterator.isNull( 4 ) ?
				RhdcHackProgress { 0, 0, 0, false } :
				RhdcHackProgress {
					(ubyte)iterator.getInt( 4 ),
					(ubyte)iterator.getInt( 5 ),
					(ushort)iterator.getInt( 6 ),
					iterator.getBool( 7 )
				}
		};
	}

	return hackMap;
}

HashMap<string,string> DataProvider::fetchAllRhdcHackVersions() {
	static SqlCommand s_cmd( R"#(
SELECT sha1, id
FROM RHDC_HACK_VERSIONS
)#"
	);

	HashMap<string,string> hackMap;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		hackMap[iterator.getString( 0 )] = iterator.getString( 1 );
	}

	return hackMap;
}

HashMap<string,StarpowerExt> DataProvider::fetchAllStarpower() {
	static SqlCommand s_cmd( R"#(
SELECT
	id,
	my_rating,
	my_difficulty,
	stars_collected,
	completed,
	MAX( play_time ),
	star_count
FROM ROMS
INNER JOIN RHDC_HACK_VERSIONS USING( sha1 )
INNER JOIN RHDC_HACKS USING( id )
INNER JOIN RHDC_HACK_PROGRESS USING( id )
GROUP BY( id )
)#"
	);

	HashMap<string,StarpowerExt> starpowerMap;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		starpowerMap[iterator.getString( 0 )] = StarpowerExt{
			RhdcHackProgress{
				(ubyte)iterator.getInt( 1 ),
				(ubyte)iterator.getInt( 2 ),
				(ushort)iterator.getInt( 3 ),
				iterator.getBool( 4 )
			},
			iterator.getLong( 5 ),
			iterator.getInt( 6 )
		};
	}

	return starpowerMap;
}

bool DataProvider::tryFetchRhdcHackByChecksum( const string &sha1, RhdcHack *hack ) noexcept {
	static SqlCommand s_cmd( R"#(
SELECT
	id,
	name,
	slug,
	description,
	star_count,
	downloads,
	avg_rating,
	avg_difficulty,
	category,
	my_rating,
	my_difficulty,
	stars_collected,
	completed
FROM RHDC_HACK_VERSIONS
INNER JOIN RHDC_HACKS USING( id )
LEFT JOIN RHDC_HACK_PROGRESS USING( id )
WHERE sha1 = ?1
)#"
	);

	s_cmd.addParameter( sha1, DATA_REF );
	DataRecordIterator iterator = s_cmd.execQuery();
	if( !iterator.moveNext() ) {
		return false;
	}

	hack->info = RhdcHackInfo{
		iterator.getString( 0 ),
		iterator.getString( 1 ),
		iterator.getString( 2 ),
		iterator.getString( 3 ),
		iterator.getInt( 4 ),
		iterator.getInt( 5 ),
		iterator.getFloat( 6 ),
		iterator.getFloat( 7 ),
		iterator.getString( 8 )
	};

	if( iterator.isNull( 9 ) ) {
		hack->progress = RhdcHackProgress{ 0, 0, 0, false };
	} else {
		hack->progress = RhdcHackProgress{
			(ubyte)iterator.getInt( 9 ),
			(ubyte)iterator.getInt( 10 ),
			(ushort)iterator.getInt( 11 ),
			iterator.getBool( 12 )
		};
	}

	return true;
}

std::set<string> DataProvider::fetchGroupsForRhdcHack( const string &hackId ) {
	static SqlCommand s_cmd( R"#(
SELECT DISTINCT group_name
FROM ROM_GROUPS INNER JOIN RHDC_HACK_VERSIONS USING( sha1 )
WHERE id = ?1
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );

	std::set<string> groups;
	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		groups.insert( iterator.getString( 0 ) );
	}
	return groups;
}

void DataProvider::addRhdcHackToGroup( const string &hackId, const string &groupName ) noexcept {
	static SqlCommand s_cmd( R"#(
INSERT OR IGNORE INTO ROM_GROUPS( sha1, group_name )
SELECT sha1, ?2 AS group_name
FROM RHDC_HACK_VERSIONS INNER JOIN ROMS USING( sha1 ) 
WHERE id = ?1
)#"
	);

	s_cmd.addParameter( hackId, DATA_REF );
	s_cmd.addParameter( groupName, DATA_REF );
	s_cmd.execNonQuery();
}

bool DataProvider::hasRhdcData() noexcept {
	static SqlCommand s_cmd( R"#(
SELECT id FROM RHDC_HACK_PROGRESS
UNION
SELECT id FROM RHDC_HACK_VERSIONS
LIMIT 1
)#"
	);

	DataRecordIterator iterator = s_cmd.execQuery();
	return iterator.moveNext();
}

void DataProvider::clearAllRhdcData() noexcept {
	static SqlBatch s_cmd( R"#(
DELETE FROM RHDC_HACKS;
DELETE FROM RHDC_HACK_PROGRESS;
DELETE FROM RHDC_HACK_VERSIONS;
)#"
	);

	s_cmd.execNonQuery();
}

static inline void deleteOldLogs() noexcept {
	static SqlCommand s_cmd( R"#(
DELETE FROM LOGS
WHERE ROWID IN (
	SELECT ROWID FROM LOGS
	ORDER BY ROWID DESC
	LIMIT -1 OFFSET 100
)
)#"
	);

	s_cmd.execNonQuery();
}

void DataProvider::addLogEntry( const LogEntry &logEntry ) noexcept {
	static int cleanupCounter = -1;
	static SqlCommand s_cmd( R"#(
INSERT INTO LOGS( level, timestamp, message )
VALUES( ?1, ?2, ?3 )
)#"
	);

	s_cmd.addParameter( (int)logEntry.level );
	s_cmd.addParameter( logEntry.timestamp );
	s_cmd.addParameter( logEntry.message, DATA_REF );
	s_cmd.execNonQuery();

	if( ++cleanupCounter % 10 == 0 ) {
		cleanupCounter = 0;
		deleteOldLogs();
	}
}

std::vector<LogEntry> DataProvider::fetchLogs() {
	static SqlCommand s_cmd( R"#(
SELECT level, timestamp, message
FROM LOGS
ORDER BY ROWID DESC
)#"
	);

	std::vector<LogEntry> logs;
	logs.reserve( 109 );

	DataRecordIterator iterator = s_cmd.execQuery();
	while( iterator.moveNext() ) {
		logs.push_back({
			(LogLevel)iterator.getInt( 0 ),
			iterator.getLong( 1 ),
			iterator.getString( 2 )
		});
	}

	return logs;
}
