#include "src/rhdc/ui/rhdc-login-dialog.hpp"
#include "ui_rhdc-login-dialog.h"

#include <QMessageBox>
#include "src/polyfill/file-dialog.hpp"
#include "src/polyfill/sha1.hpp"
#include "src/core/preset-controllers.hpp"
#include "src/core/file-controller.hpp"
#include "src/core/sm64.hpp"
#include "src/core/rom.hpp"
#include "src/db/data-provider.hpp"
#include "src/ui/icons.hpp"
#include "src/rhdc/web/api.hpp"

RhdcLoginDialog::RhdcLoginDialog( QWidget *parent ) :
	QDialog( parent ),
	m_ui( new Ui::RhdcLoginDialog ),
	m_exists( new bool( true ) )
{
	m_ui->setupUi( this );
	m_ui->settingsContainer->setVisible( false );
	m_ui->browseForRomButton->setIcon( Icon::browse() );
	m_ui->downloadDirButton->setIcon( Icon::browse() );
	m_ui->loginButton->setIcon( Icon::login() );
	setWindowIcon( Icon::appIcon() );
}

RhdcLoginDialog::~RhdcLoginDialog() {
	*m_exists = false;
	delete m_ui;
}

void RhdcLoginDialog::showEvent( QShowEvent *event ) {
	fs::path _discard;
	m_ui->pages->setCurrentIndex( SM64::tryGetPath( _discard ) ? 1 : 0 );
	m_ui->usernameInput->clear();
	m_ui->passwordInput->clear();
	m_ui->loginErrorMessage->clear();
	m_ui->buttonTray->button( QDialogButtonBox::Cancel )->setVisible( true );
	m_ui->buttonTray->button( QDialogButtonBox::Ok )->setVisible( false );
	m_ui->loginSpinner->setVisible( false );
	m_ui->loginButton->setEnabled( true );
	m_ui->settingsContainer->setVisible( false );
	QDialog::showEvent( event );

	QRect geo = geometry();
	geo.setWidth( 450 );
	geo.setHeight( 275 );
	setGeometry( geo );
}

static const string SM64_SHA1 = "9bef1128717f958171a4afac3ed78ee2bb4e86ce";

void RhdcLoginDialog::browseForRom() {
	const fs::path romPath = FileDialog::getFile( "Select SM64 ROM", {{ "z64 ROM", "*.z64" }} );
	if( romPath.empty() ) return;

	if( Sha1::compute( romPath ) != SM64_SHA1 ) {
		QMessageBox::critical( this, "ROM Does not Match", "The provided rom does not match the expected checksum." );
		return;
	}

	DataProvider::addManualRomPath( romPath );
	DataProvider::addRomFile({
		romPath,
		RomUtil::getLastModified( romPath ),
		SM64_SHA1
	});
	DataProvider::addRomInfo({
		SM64_SHA1,
		"Super Mario 64",
		"SUPER MARIO 64",
		EmulatorCore::ParallelN64,
		GfxPlugin::ParaLLEl,
		GfxPlugin::ParaLLEl,
		true,
		false,
		true,
		0,
		0,
		std::set<string>(),
		true,
		false,
		1021708041,
		DefaultInputModes::Normal.id,
		false
	});

	m_ui->pages->setCurrentIndex( 1 );
}

void RhdcLoginDialog::browseForDir() {
	const fs::path downloadDir = FileDialog::getDirectory( "Select a new download directory" );
	if( downloadDir.empty() ) return;
	m_ui->downloadDirInput->setText( downloadDir.u8string().c_str() );
}

void RhdcLoginDialog::login() {
	if( m_ui->usernameInput->text().isEmpty() ) {
		m_ui->loginErrorMessage->setText( "Enter your username" );
		return;
	}

	if( m_ui->passwordInput->text().isEmpty() ) {
		m_ui->loginErrorMessage->setText( "Enter your password" );
		return;
	}

	m_ui->loginButton->setEnabled( false );
	m_ui->loginSpinner->setVisible( true );
	m_ui->loginErrorMessage->clear();

	const RhdcCredentials credentials = {
		m_ui->usernameInput->text().toStdString(),
		m_ui->passwordInput->text().toStdString()
	};

	std::shared_ptr<bool> dialogExists = m_exists;
	RhdcApi::loginAsync( credentials,
		[this,dialogExists,credentials]() {
			credentials.save();
			if( !*dialogExists ) return;
			this->m_ui->pages->setCurrentIndex( 2 );
			this->m_ui->buttonTray->button( QDialogButtonBox::Cancel )->setVisible( false );
			this->m_ui->buttonTray->button( QDialogButtonBox::Ok )->setVisible( true );

			QRect geo = this->geometry();
			geo.setWidth( 525 );
			geo.setHeight( 575 );
			this->setGeometry( geo );
			this->m_ui->settingsContainer->setVisible( true );

			const RhdcSettings rhdcSettings = FileController::loadRhdcSettings();
			this->m_ui->downloadDirInput->setText( rhdcSettings.downloadDirectory.u8string().c_str() );
			this->m_ui->allSlotsCheckbox->setChecked( rhdcSettings.checkAllSaveSlots );
			this->m_ui->preferHleCheckbox->setChecked( rhdcSettings.preferHle );
			this->m_ui->syncGroupsCheckbox->setChecked( rhdcSettings.syncGroups );
		},
		[this,dialogExists](ApiErrorType error) {
			if( !*dialogExists ) return;
			this->m_ui->loginSpinner->setVisible( false );
			this->m_ui->loginButton->setEnabled( true );
			switch( error ) {
				case ApiErrorType::NotAuthorized:
					this->m_ui->loginErrorMessage->setText( "Username or password is incorrect" );
					break;
				case ApiErrorType::RegistrationRequired:
					this->m_ui->loginErrorMessage->setText( "This account requires e-mail confirmation to activate." );
					break;
				case ApiErrorType::ServerDown:
					this->m_ui->loginErrorMessage->setText( "The romhacking.com servers appear to be down." );
					break;
				default:
					this->m_ui->loginErrorMessage->setText( "An unknown error occurred." );
					break;
			}
		}
	);
}

void RhdcLoginDialog::accept() {
	FileController::saveRhdcSettings({
		fs::to_path( m_ui->downloadDirInput->text().toStdString() ),
		m_ui->allSlotsCheckbox->isChecked(),
		m_ui->preferHleCheckbox->isChecked(),
		m_ui->syncGroupsCheckbox->isChecked()
	});
	QDialog::accept();
}
